"""
  Directs the executor to include this field or
  fragment only when the `if` argument is true.
"""
directive @include("Included when true." if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT

"""
  Directs the executor to skip this field or
  fragment when the `if` argument is true.
"""
directive @skip("Skipped when true." if: Boolean!) on FIELD | FRAGMENT_SPREAD | INLINE_FRAGMENT
